"""Collection of General Projective-Geometry Functions"""

# global
import ivy
from operator import mul
from functools import reduce


def transform(coords, trans, batch_shape=None, image_shape=None):
    r"""Transform image of :math:`n`-dimensional co-ordinates :math:`\mathbf{
    x}\in\mathbb{R}^{h×w×n}` by transformation matrix :math:`\mathbf{f}\in\mathbb{
    R}^{m×n}`, to produce image of transformed co-ordinates :math:`\mathbf{x}_{
    trans}\in\mathbb{R}^{h×w×m}`.\n `[reference]
    <https://en.wikipedia.org/wiki/Matrix_multiplication>`_

    Parameters
    ----------
    coords
        Co-ordinate image *[batch_shape,height,width,n]*
    trans
        Transformation matrix *[batch_shape,m,n]*
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_shape
        Image shape. Inferred from inputs if None. (Default value = None)

    Returns
    -------
    ret
        Transformed co-ordinate image *[batch_shape,height,width,m]*

    """
    if batch_shape is None:
        batch_shape = trans.shape[:-2]
    num_batch_dims = len(batch_shape)

    if image_shape is None:
        image_shape = coords.shape[num_batch_dims:-1]
    image_shape_flat = reduce(mul, image_shape)

    # shapes as list
    batch_shape = list(batch_shape)
    image_shape = list(image_shape)

    # BS x ISF x N
    coords_flattened = ivy.reshape(coords, batch_shape + [image_shape_flat, -1])

    # BS x N x ISF
    coords_reshaped = ivy.swapaxes(coords_flattened, -1, -2)

    # BS x M x ISF
    transformed_coords_vector = ivy.matmul(trans, coords_reshaped)

    # BS x ISF x M
    transformed_coords_vector_transposed = ivy.swapaxes(
        transformed_coords_vector, -1, -2
    )

    # BS x IS x M
    return ivy.reshape(
        transformed_coords_vector_transposed, batch_shape + image_shape + [-1]
    )


def projection_matrix_pseudo_inverse(proj_mat, batch_shape=None):
    r"""Given projection matrix :math:`\mathbf{P}\in\mathbb{R}^{3×4}`, compute it's
    pseudo-inverse :math:`\mathbf{P}^+\in\mathbb{R}^{4×3}`.\n `[reference]
    <localhost:63342/ivy/docs/source/references/mvg_textbook.pdf?_ijt
    =25ihpil89dmfo4da975v402ogc#page=179>`_ bottom of page 161, section 6.2.2

    Parameters
    ----------
    proj_mat
        Projection matrix *[batch_shape,3,4]*
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)

    Returns
    -------
    ret
        Projection matrix pseudo-inverse *[batch_shape,4,3]*

    """
    if batch_shape is None:
        batch_shape = proj_mat.shape[:-2]

    # shapes as list
    batch_shape = list(batch_shape)

    # transpose idxs
    num_batch_dims = len(batch_shape)
    transpose_idxs = list(range(num_batch_dims)) + [num_batch_dims + 1, num_batch_dims]

    # BS x 4 x 3
    matrix_transposed = ivy.permute_dims(proj_mat, axes=transpose_idxs)

    # BS x 4 x 3
    return ivy.matmul(
        matrix_transposed, ivy.inv(ivy.matmul(proj_mat, matrix_transposed))
    )


def projection_matrix_inverse(proj_mat):
    r"""Given projection matrix :math:`\mathbf{P}\in\mathbb{R}^{3×4}`, compute it's
    inverse :math:`\mathbf{P}^{-1}\in\mathbb{R}^{3×4}`.\n `[reference]
    <https://github.com/pranjals16/cs676/blob/master/Hartley%2C%20Zisserman%20
    -%20Multiple%20View%20Geometry%20in%20Computer%20Vision.pdf#page=174>`_ middle of
    page 156, section 6.1, eq 6.6

    Parameters
    ----------
    proj_mat
        Projection matrix *[batch_shape,3,4]*

    Returns
    -------
    ret
        Projection matrix inverse *[batch_shape,3,4]*

    """
    # BS x 3 x 3
    rotation_matrix = proj_mat[..., 0:3]

    # BS x 3 x 3
    rotation_matrix_inverses = ivy.inv(rotation_matrix)

    # BS x 3 x 1
    translations = proj_mat[..., 3:4]

    # BS x 3 x 1
    translation_inverses = -ivy.matmul(rotation_matrix_inverses, translations)

    # BS x 3 x 4
    return ivy.concat((rotation_matrix_inverses, translation_inverses), axis=-1)


def solve_homogeneous_dlt(A):
    r"""Given :math:`\mathbf{A}\in\mathbb{R}^{d×d}`, solve the system of :math:`d`
    equations in :math:`d` unknowns :math:`\mathbf{Ax} = \mathbf{0}` using the
    homogeneous DLT method, to return :math:`\mathbf{x}\in\mathbb{R}^d`.\n `[
    reference] <https://github.com/pranjals16/cs676/blob/master/Hartley%2C
    %20Zisserman%20-%20Multiple%20View%20Geometry%20in%20Computer%20Vision.pdf#page
    =106>`_ bottom of page 88, section 4.1

    Parameters
    ----------
    A
        Matrix representing system of equations to solve *[batch_shape,d,d]*

    Returns
    -------
    ret
        Solution to the system of equations *[batch_shape,d]*

    """
    # BS x D x D,    BS x D,    BS x D x D
    U, D, VT = ivy.svd(A)

    # BS x D
    return VT[..., -1, :]



"""Collection of Voxel-Grid Functions"""

# global
import ivy

MIN_DENOMINATOR = 1e-12


def coords_to_voxel_grid(
    coords,
    voxel_shape_spec,
    mode="DIMS",
    coord_bounds=None,
    features=None,
    batch_shape=None,
    device=None,
):
    r"""Create voxel grid :math:`\mathbf{X}_v\in\mathbb{R}^{x×y×z×(3+N+1)}` from
    homogeneous co-ordinates :math:`\mathbf{X}_w\in\mathbb{R}^{num\_coords×4}`. Each
    voxel contains 3+N+1 values: the mean world co-ordinate inside the voxel for the
    projected pixels, N coordinte features (optional), and also the number of
    projected pixels inside the voxel. Grid resolutions and dimensions are also
    returned separately for each entry in the batch. Note that the final batched
    voxel grid returned uses the maximum grid dimensions across the batch, this means
    some returned grids may contain redundant space, with all but the single largest
    batched grid occupying a subset of the grid space, originating from the corner of
    minimum :math:`x,y,z` values.\n `[reference]
    <https://en.wikipedia.org/wiki/Voxel>`_

    Parameters
    ----------
    coords
        Homogeneous co-ordinates *[batch_shape,c,4]*
    voxel_shape_spec
        Either the number of voxels in x,y,z directions, or the resolutions (metres) in
        x,y,z
        directions, depending on mode. Batched or unbatched. *[batch_shape,3]* or *[3]*
    mode
        Shape specification mode, either "DIMS" or "RES" (Default value = 'DIMS')
    coord_bounds
        Co-ordinate x, y, z boundaries *[batch_shape,6]* or *[6]* (Default value = None)
    features
        Co-ordinate features *[batch_shape,c,4]*.
        E.g. RGB values, low-dimensional features, etc.
        Features mapping to the same voxel are averaged. (Default value = None)
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    device
        device on which to create the array 'cuda:0', 'cuda:1', 'cpu' etc.
        Same as x if None. (Default value = None)

    Returns
    -------
    ret
        Voxel grid *[batch_shape,x_max,v_max,z_max,3+feature_size+1]*, dimensions
        *[batch_shape,3]*,
        resolutions *[batch_shape,3]*, voxel_grid_lower_corners *[batch_shape,3]*

    """
    if batch_shape is None:
        batch_shape = coords.shape[:-2]

    if device is None:
        device = ivy.dev(coords)

    # shapes as list
    batch_shape = list(batch_shape)
    num_batch_dims = len(batch_shape)
    num_coords_per_batch = coords.shape[-2]

    # voxel shape spec as array
    if len(voxel_shape_spec) == 3:
        # BS x 1 x 3
        voxel_shape_spec = ivy.expand_dims(
            ivy.tile(
                ivy.reshape(ivy.array(voxel_shape_spec), [1] * num_batch_dims + [3]),
                batch_shape + [1],
            ),
            axis=-2,
        )

    # coord bounds spec as array
    if coord_bounds is not None:
        if len(coord_bounds) == 6:
            # BS x 6
            coord_bounds = ivy.tile(
                ivy.reshape(
                    ivy.array(coord_bounds, dtype="float32"), [1] * num_batch_dims + [6]
                ),
                batch_shape + [1],
            )

        # BS x 1 x 6
        coord_bounds = ivy.expand_dims(coord_bounds, axis=-2)

    # BS x N x 3
    coords = coords[..., 0:3]

    if coord_bounds is not None:
        # BS x 1 x 1
        x_min = coord_bounds[..., 0:1]
        y_min = coord_bounds[..., 1:2]
        z_min = coord_bounds[..., 2:3]
        x_max = coord_bounds[..., 3:4]
        y_max = coord_bounds[..., 4:5]
        z_max = coord_bounds[..., 5:6]

        # BS x N x 1
        x_coords = coords[..., 0:1]
        y_coords = coords[..., 1:2]
        z_coords = coords[..., 2:3]

        x_validity_mask = ivy.logical_and(x_coords > x_min, x_coords < x_max)
        y_validity_mask = ivy.logical_and(y_coords > y_min, y_coords < y_max)
        z_validity_mask = ivy.logical_and(z_coords > z_min, z_coords < z_max)

        # BS x N
        full_validity_mask = ivy.logical_and(
            ivy.logical_and(x_validity_mask, y_validity_mask), z_validity_mask
        )[..., 0]

        # BS x 1 x 3
        bb_mins = coord_bounds[..., 0:3]
        bb_maxs = coord_bounds[..., 3:6]
        bb_ranges = bb_maxs - bb_mins
    else:
        # BS x N
        full_validity_mask = ivy.astype(
            ivy.ones(batch_shape + [num_coords_per_batch], device=device), "bool"
        )

        # BS x 1 x 3
        bb_mins = ivy.min(coords, axis=-2, keepdims=True)
        bb_maxs = ivy.max(coords, axis=-2, keepdims=True)
        bb_ranges = bb_maxs - bb_mins

    # get voxel dimensions
    if mode == "DIMS":
        # BS x 1 x 3
        dims = ivy.astype(voxel_shape_spec, "int32")
    elif mode == "RES":
        # BS x 1 x 3
        res = ivy.astype(voxel_shape_spec, "float32")
        dims = ivy.astype(ivy.ceil(bb_ranges / (res + MIN_DENOMINATOR)), "int32")
    else:
        raise Exception('Invalid mode selection. Must be either "DIMS" or "RES"')
    dims_m_one = ivy.astype(dims - 1, "int32")

    # BS x 1 x 3
    res = bb_ranges / (ivy.astype(dims, "float32") + MIN_DENOMINATOR)

    # BS x NC x 3
    voxel_indices = ivy.minimum(
        ivy.astype(ivy.floor((coords - bb_mins) / (res + MIN_DENOMINATOR)), "int32"),
        dims_m_one,
    )

    # BS x NC x 3
    voxel_values = coords

    feature_size = 0
    if features is not None:
        feature_size = features.shape[-1]
        voxel_values = ivy.concat([voxel_values, features], axis=-1)

    # TNVC x len(BS)+1
    valid_coord_indices = ivy.astype(ivy.argwhere(full_validity_mask), "int32")

    # scalar
    total_num_valid_coords = valid_coord_indices.shape[0]

    # TNVC x 3
    voxel_values_pruned_flat = ivy.gather_nd(voxel_values, valid_coord_indices)
    voxel_indices_pruned_flat = ivy.gather_nd(voxel_indices, valid_coord_indices)

    # TNVC x len(BS)+2
    if num_batch_dims == 0:
        all_indices_pruned_flat = voxel_indices_pruned_flat
    else:
        batch_indices = valid_coord_indices[..., :-1]
        all_indices_pruned_flat = ivy.concat(
            [batch_indices] + [voxel_indices_pruned_flat], axis=-1
        )

    # TNVC x 4
    voxel_values_pruned_flat = ivy.concat(
        (
            voxel_values_pruned_flat,
            ivy.ones([total_num_valid_coords, 1], device=device),
        ),
        axis=-1,
    )

    # get max dims list for scatter
    if num_batch_dims > 0:
        max_dims = ivy.max(
            ivy.reshape(dims, batch_shape + [3]), axis=list(range(num_batch_dims))
        )
    else:
        max_dims = ivy.reshape(dims, batch_shape + [3])
    batch_shape_array_list = (
        [ivy.array(batch_shape, dtype="int32", device=device)]
        if num_batch_dims != 0
        else []
    )
    total_dims_list = ivy.to_list(
        ivy.concat(
            batch_shape_array_list
            + [max_dims, ivy.array([4 + feature_size], dtype="int32", device=device)],
            axis=-1,
        )
    )

    # BS x x_max x y_max x z_max x 4
    scattered = ivy.scatter_nd(
        all_indices_pruned_flat,
        voxel_values_pruned_flat,
        shape=total_dims_list,
        reduction="sum",
    )

    # BS x x_max x y_max x z_max x 4 + feature_size, BS x 3, BS x 3, BS x 3
    return (
        ivy.concat(
            (
                scattered[..., :-1]
                / (ivy.maximum(scattered[..., -1:], 1.0) + MIN_DENOMINATOR),
                scattered[..., -1:],
            ),
            axis=-1,
        ),
        dims[..., 0, :],
        res[..., 0, :],
        bb_mins[..., 0, :],
    )



# lint as: python3
# Copyright 2021 The Ivy Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License..
# ==============================================================================
import setuptools
from pathlib import Path
from distutils.core import setup


def _strip(line):
    return line.split(" ")[0].split("#")[0].split(",")[0]


def _replace_logos_html(txt):
    # html-containing chunks
    chunks = txt.split(".. raw:: html")

    # backend logos
    backends_chunk = chunks[2]
    bc = backends_chunk.split("\n\n")
    img_str = (
        ".. image:: https://github.com/unifyai/unifyai.github.io/blob/master/img/externally_linked/logos/supported/frameworks.png?raw=true\n"  # noqa
        "   :width: 100%"
    )
    backends_chunk = "\n\n".join(bc[0:1] + [img_str] + bc[2:])

    # library logos
    libraries_chunk = chunks[3]
    lc = libraries_chunk.split("\n\n")
    img_str = (
        ".. image:: https://github.com/unifyai/unifyai.github.io/blob/master/img/externally_linked/ivy_libraries.png?raw=true\n"  # noqa
        "   :width: 100%"
    )
    libraries_chunk = "\n\n".join(lc[0:1] + [img_str] + lc[2:])

    # re-join
    chunks[3] = libraries_chunk
    return "".join(
        [
            ".. raw:: html".join(chunks[0:2]),
            backends_chunk,
            ".. raw:: html".join(chunks[3:]),
        ]
    )


def _replace_gif(gif_chunk):
    png_url = "https://{}.png".format(
        gif_chunk.split(".gif?raw=true'>")[0].split("https://")[-1]
    )
    gc = gif_chunk.split("\n\n")
    img_str = ".. image:: {}?raw=true\n" "   :width: 100%".format(png_url)
    return "\n\n".join(gc[0:1] + [img_str] + gc[2:])


def _replace_gifs_html(txt):
    # html-containing chunks
    chunks = txt.split(".. raw:: html")

    # go through each chunk, replacing all html gifs with rst images
    return_str = ""
    for i, chunk in enumerate(chunks):
        new_chunk = chunk
        delimiter = ".. raw:: html"
        if ".gif?raw=true'>" in chunk:
            new_chunk = _replace_gif(chunk)
            delimiter = ""
        if i == 0:
            return_str = chunk
        else:
            return_str = delimiter.join([return_str, new_chunk])
    return return_str


def _is_html(line):
    line_squashed = line.replace(" ", "")
    if not line_squashed:
        return False
    if line_squashed[0] == "<" and line_squashed[-1] == ">":
        return True
    return False


def _is_raw_block(line):
    line_squashed = line.replace(" ", "")
    if len(line_squashed) < 11:
        return False
    if line_squashed[-11:] == "..raw::html":
        return True
    return False


this_directory = Path(__file__).parent
text = (this_directory / "README.rst").read_text()
text = _replace_logos_html(text).replace(
    ". Click on the icons below to learn more!", "!"
)
text = _replace_gifs_html(text)
lines = text.split("\n")
lines = [line for line in lines if not (_is_html(line) or _is_raw_block(line))]
long_description = "\n".join(lines)

setup(
    name="ivy-vision",
    version="0.0.0.dev0",
    author="Unify",
    author_email="hello@unify.ai",
    description=(
        "3D Vision functions with end-to-end support for "
        "machine learning developers, written in Ivy."
    ),
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://unify.ai/docs/vision",
    project_urls={
        "Docs": "https://unify.ai/docs/vision/",
        "Source": "https://github.com/unifyai/vision",
    },
    packages=setuptools.find_packages(),
    install_requires=[_strip(line) for line in open("requirements.txt", "r")],
    classifiers=["License :: OSI Approved :: Apache Software License"],
    license="Apache 2.0",
)



"""Collection of Image Functions"""

# global
import ivy
import numpy as np
import operator
import functools
import math


def stack_images(images, desired_aspect_ratio):
    """Stacks a group of images into a combined windowed image, fitting the desired
    aspect ratio as closely as possible.

    Parameters
    ----------
    images
        Sequence of image arrays to be stacked *[batch_shape,height,width,dims]*
    desired_aspect_ratio:
        desired aspect ratio of the stacked image
    Returns
    -------
    ret
        an array containing the stacked images in a specified aspect ratio/dimensions

    """
    num_images = len(images)
    if num_images == 0:
        raise Exception("At least 1 image must be provided")
    batch_shape = ivy.shape(images[0])[:-3]
    image_dims = ivy.shape(images[0])[-3:-1]
    num_batch_dims = len(batch_shape)
    if num_images == 1:
        return images[0]
    img_ratio = image_dims[0] / image_dims[1]
    desired_img_ratio = desired_aspect_ratio[0] / desired_aspect_ratio[1]
    stack_ratio = img_ratio * desired_img_ratio
    stack_height = (num_images / stack_ratio) ** 0.5
    stack_height_int = math.ceil(stack_height)
    stack_width_int = math.ceil(num_images / stack_height)
    image_rows = list()
    for i in range(stack_width_int):
        images_to_concat = images[i * stack_height_int : (i + 1) * stack_height_int]
        images_to_concat += [
            ivy.zeros_like(
                images[0], dtype=ivy.dtype(images[0]), device=ivy.dev(images[0])
            )
        ] * (stack_height_int - len(images_to_concat))
        image_rows.append(ivy.concat(images_to_concat, num_batch_dims))

    return ivy.concat(image_rows, axis=num_batch_dims + 1)


def gradient_image(x):
    """Computes image gradients (dy, dx) for each channel.

    Parameters
    ----------
    x
        Input image *[batch_shape, h, w, d]* .

    Returns
    -------
    ret
        Gradient images dy *[batch_shape,h,w,d]* and dx *[batch_shape,h,w,d]* .

    """
    x_shape = ivy.shape(x)
    batch_shape = x_shape[:-3]
    image_dims = x_shape[-3:-1]
    device = ivy.dev(x)
    # to list
    batch_shape = list(batch_shape)
    image_dims = list(image_dims)
    num_dims = x_shape[-1]
    # BS x H-1 x W x D
    dy = x[..., 1:, :, :] - x[..., :-1, :, :]
    # BS x H x W-1 x D
    dx = x[..., :, 1:, :] - x[..., :, :-1, :]
    # BS x H x W x D
    dy = ivy.concat(
        (dy, ivy.zeros(batch_shape + [1, image_dims[1], num_dims], device=device)),
        axis=-3,
    )
    dx = ivy.concat(
        (dx, ivy.zeros(batch_shape + [image_dims[0], 1, num_dims], device=device)),
        axis=-2,
    )
    # BS x H x W x D,    BS x H x W x D
    return dx, dy


def float_img_to_uint8_img(x):
    """Converts an image of floats into a bit-cast 4-channel image of uint8s, which can
    be saved to disk.

    Parameters
    ----------
    x
        Input float image *[batch_shape,h,w]*.

    Returns
    -------
    ret
        The new encoded uint8 image *[batch_shape,h,w,4]* .
    """
    x_np = ivy.to_numpy(x).astype("float32")
    x_shape = x_np.shape
    x_bytes = x_np.tobytes()
    x_uint8 = np.frombuffer(x_bytes, np.uint8)
    return ivy.array(np.reshape(x_uint8, list(x_shape) + [4]).tolist())


def uint8_img_to_float_img(x):
    """Converts an image of uint8 values into a bit-cast float image.

    Parameters
    ----------
    x
        Input uint8 image *[batch_shape,h,w,4]*

    Returns
    -------
    ret
        The new float image *[batch_shape,h,w]*

    """
    x_np = ivy.to_numpy(x).astype("uint8")
    x_shape = x_np.shape
    x_bytes = x_np.tobytes()
    x_float = np.frombuffer(x_bytes, np.float32)
    return ivy.array(np.reshape(x_float, x_shape[:-1]).tolist())


def random_crop(x, crop_size, batch_shape=None, image_dims=None, seed: int = None):
    """Randomly crops the input images.

    Parameters
    ----------
    x
        Input images to crop *[batch_shape,h,w,f]*
    crop_size
        The 2D crop size.
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_dims
        Image dimensions. Inferred from inputs in None. (Default value = None)
    seed
        Required for random number generator

    Returns
    -------
    ret
        The new cropped image *[batch_shape,nh,nw,f]*
    """
    x_shape = x.shape
    if batch_shape is None:
        batch_shape = x_shape[:-3]
    if image_dims is None:
        image_dims = x_shape[-3:-1]
    num_channels = x_shape[-1]
    flat_batch_size = functools.reduce(operator.mul, [batch_shape], 1)
    crop_size[0] = min(crop_size[-2], x_shape[-3])
    crop_size[1] = min(crop_size[-1], x_shape[-2])

    # shapes as list
    image_dims = list(image_dims)
    margins = [img_dim - cs for img_dim, cs in zip(image_dims, crop_size)]

    # FBS x H x W x F
    x_flat = ivy.reshape(x, [flat_batch_size] + image_dims + [num_channels])

    # FBS x 1
    rng = np.random.default_rng(seed)
    x_offsets = rng.integers(0, margins[0] + 1, flat_batch_size).tolist()
    y_offsets = rng.integers(0, margins[1] + 1, flat_batch_size).tolist()

    # list of 1 x NH x NW x F
    cropped_list = [
        img[..., xo : xo + crop_size[0], yo : yo + crop_size[1], :]
        for img, xo, yo in zip(
            ivy.unstack(x_flat, axis=0, keepdims=True), x_offsets, y_offsets
        )
    ]

    # FBS x NH x NW x F
    flat_cropped = ivy.concat(cropped_list, axis=0)

    # BS x NH x NW x F
    return ivy.reshape(flat_cropped, [batch_shape] + crop_size + [num_channels])


def bilinear_resample(x, warp):
    """Performs bilinearly re-sampling on input image.

    Parameters
    ----------
    x
        Input image *[batch_shape,h,w,dims]*.
    warp
        Warp array *[batch_shape,num_samples,2]*

    Returns
    -------
    ret
        Image after bilinear re-sampling.

    """
    batch_shape = x.shape[:-3]
    input_image_dims = x.shape[-3:-1]
    num_feats = x.shape[-1]
    batch_shape = list(batch_shape)
    input_image_dims = list(input_image_dims)
    # image statistics
    height, width = input_image_dims
    max_x = width - 1
    max_y = height - 1
    warp.shape[-2]
    batch_shape_flat = int(ivy.prod(ivy.asarray(batch_shape)))
    # B
    batch_offsets = ivy.arange(batch_shape_flat) * height * width
    # B x (HxW)
    base_grid = ivy.tile(ivy.expand_dims(batch_offsets, axis=1), [1, height * width])
    # (BxHxW)
    base = ivy.reshape(base_grid, [-1])
    # (BxHxW) x D
    data_flat = ivy.reshape(x, [batch_shape_flat * height * width, -1])
    # (BxHxW) x 2
    warp_flat = ivy.reshape(warp, [-1, 2])
    warp_floored = (ivy.floor(warp_flat)).astype("int32")
    bilinear_weights = warp_flat - ivy.floor(warp_flat)
    # (BxHxW)
    x0 = warp_floored[:, 0]
    x1 = x0 + 1
    y0 = warp_floored[:, 1]
    y1 = y0 + 1
    x0 = ivy.clip(x0, 0, max_x)
    x1 = ivy.clip(x1, 0, max_x)
    y0 = ivy.clip(y0, 0, max_y)
    y1 = ivy.clip(y1, 0, max_y)
    base_y0 = base + y0 * width
    base_y1 = base + y1 * width
    idx_a = base_y0 + x0
    idx_b = base_y1 + x0
    idx_c = base_y0 + x1
    idx_d = base_y1 + x1

    # (BxHxW) x D
    Ia = ivy.gather(data_flat, ivy.reshape(idx_a, [-1, 1]), axis=0).squeeze(axis=1)
    Ib = ivy.gather(data_flat, ivy.reshape(idx_b, [-1, 1]), axis=0).squeeze(axis=1)
    Ic = ivy.gather(data_flat, ivy.reshape(idx_c, [-1, 1]), axis=0).squeeze(axis=1)
    Id = ivy.gather(data_flat, ivy.reshape(idx_d, [-1, 1]), axis=0).squeeze(axis=1)

    # (BxHxW)
    xw = bilinear_weights[:, 0]
    yw = bilinear_weights[:, 1]
    # (BxHxW) x 1
    wa = ivy.expand_dims((1 - xw) * (1 - yw), axis=1)
    wb = ivy.expand_dims((1 - xw) * yw, axis=1)
    wc = ivy.expand_dims(xw * (1 - yw), axis=1)
    wd = ivy.expand_dims(xw * yw, axis=1)
    # (BxNP) x D
    resampled_flat = wa * Ia + wb * Ib + wc * Ic + wd * Id
    # B x NP x D
    return ivy.reshape(resampled_flat, batch_shape + [-1, num_feats])



"""Collection of Optical Flow Functions"""

# global
import ivy
import ivy_mech

# local
from ivy_vision import image as ivy_img
from ivy_vision import two_view_geometry as ivy_tvg
from ivy_vision import projective_geometry as ivy_pg
from ivy_vision import single_view_geometry as ivy_svg

MIN_DENOMINATOR = 1e-12


def depth_from_flow_and_cam_mats(
    flow,
    full_mats,
    inv_full_mats=None,
    camera_centers=None,
    uniform_pixel_coords=None,
    triangulation_method="cmp",
    batch_shape=None,
    image_dims=None,
    device=None,
):
    r"""Compute depth map :math:`\mathbf{X}\in\mathbb{R}^{h×w×1}` in frame 1 using
    optical flow :math:`\mathbf{U}_{1→2}\in\mathbb{R}^{h×w×2}` from frame 1 to 2,
    and the camera geometry.\n

    Parameters
    ----------
    flow
        Optical flow from frame 1 to 2 *[batch_shape,h,w,2]*
    full_mats
        Full projection matrices *[batch_shape,2,3,4]*
    inv_full_mats
        Inverse full projection matrices, inferred from full_mats if None and
        'cmp' triangulation method *[batch_shape,2,3,4]* (Default value = None)
    camera_centers
        Camera centers, inferred from inv_full_mats if None and
        'cmp' triangulation method *[batch_shape,2,3,1]* (Default value = None)
    uniform_pixel_coords
        Homogeneous uniform (integer) pixel co-ordinate images,
        inferred from image_dims if None *[batch_shape,h,w,3]* (Default value = None)
    triangulation_method
        Triangulation method, one of [cmp|dlt], for closest mutual points or
        homogeneous dlt approach, closest_mutual_points by default
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_dims
        Image dimensions. Inferred from inputs in None. (Default value = None)
    device
        device on which to create the array 'cuda:0', 'cuda:1', 'cpu' etc.
        Same as x if None. (Default value = None)

    Returns
    -------
    ret
        Depth map in frame 1 *[batch_shape,h,w,1]*

    """
    if batch_shape is None:
        batch_shape = flow.shape[:-3]

    if image_dims is None:
        image_dims = flow.shape[-3:-1]

    # shapes as list
    batch_shape = list(batch_shape)
    image_dims = list(image_dims)

    if device is None:
        device = ivy.dev(flow)

    if inv_full_mats is None:
        inv_full_mats = ivy.inv(
            ivy_mech.make_transformation_homogeneous(
                full_mats, batch_shape + [2], device
            )
        )[..., 0:3, :]

    if camera_centers is None:
        camera_centers = ivy_svg.inv_ext_mat_to_camera_center(inv_full_mats)

    if uniform_pixel_coords is None:
        uniform_pixel_coords = ivy_svg.create_uniform_pixel_coords_image(
            image_dims, batch_shape, device=device
        )

    # BS x H x W x 3
    flow_homo = ivy.concat(
        (flow, ivy.zeros(batch_shape + image_dims + [1], device=device)), axis=-1
    )

    # BS x H x W x 3
    transformed_pixel_coords = uniform_pixel_coords + flow_homo

    # BS x 2 x H x W x 3
    pixel_coords = ivy.concat(
        (
            ivy.expand_dims(uniform_pixel_coords, axis=-4),
            ivy.expand_dims(transformed_pixel_coords, axis=-4),
        ),
        axis=-4,
    )

    # BS x H x W x 1
    return ivy_tvg.triangulate_depth(
        pixel_coords,
        full_mats,
        inv_full_mats,
        camera_centers,
        triangulation_method,
        batch_shape,
        image_dims,
    )[..., -1:]


def flow_from_depth_and_cam_mats(
    pixel_coords1, cam1to2_full_mat, batch_shape=None, image_shape=None
):
    r"""Compute optical flow :math:`\mathbf{U}_{1→2}\in\mathbb{R}^{is×2}`
    from frame 1 to 2, using depth map
    :math:`\mathbf{X}\in\mathbb{R}^{is×1}` in frame 1, and the camera geometry.\n

    Parameters
    ----------
    pixel_coords1
        Depth scaled homogeneous pixel co-ordinates image in frame 1
        *[batch_shape,image_shape,3]*
    cam1to2_full_mat
        Camera1-to-camera2 full projection matrix *[batch_shape,3,4]*
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_shape
        Image dimensions. Inferred from inputs in None. (Default value = None)

    Returns
    -------
    ret
        Optical flow from frame 1 to 2 *[batch_shape,image_shape,2]*

    """
    if batch_shape is None:
        batch_shape = cam1to2_full_mat.shape[:-2]
    num_batch_dims = len(batch_shape)

    if image_shape is None:
        image_shape = pixel_coords1.shape[num_batch_dims:-1]

    # shapes as list
    batch_shape = list(batch_shape)
    image_shape = list(image_shape)

    # BS x IS x 3
    projected_pixel_coords = ivy_tvg.ds_pixel_to_ds_pixel_coords(
        pixel_coords1, cam1to2_full_mat, batch_shape, image_shape
    )
    projected_pixel_coords_normalized = projected_pixel_coords / (
        projected_pixel_coords[..., -1:] + MIN_DENOMINATOR
    )

    # BS x IS x 3
    pixel_coords1_normalized = pixel_coords1 / (
        pixel_coords1[..., -1:] + MIN_DENOMINATOR
    )

    # BS x IS x 2
    return (
        projected_pixel_coords_normalized[..., 0:2] - pixel_coords1_normalized[..., 0:2]
    )


def project_flow_to_epipolar_line(
    flow,
    fund_mat,
    uniform_pixel_coords=None,
    batch_shape=None,
    image_dims=None,
    device=None,
):
    r"""Project optical flow :math:`\mathbf{U}_{1→2}\in\mathbb{R}^{h×w×2}` to epipolar
    line in frame 1.\n `[reference]
    <https://en.wikipedia.org/wiki/Distance_from_a_point_to_a_line
    #Line_defined_by_an_equation>`_

    Parameters
    ----------
    flow
        Optical flow from frame 1 to 2 *[batch_shape,h,w,2]*
    fund_mat
        Fundamental matrix connecting frames 1 and 2 *[batch_shape,3,3]*
    uniform_pixel_coords
        Homogeneous uniform (integer) pixel co-ordinate images,
        inferred from image_dims if None *[batch_shape,h,w,3]* (Default value = None)
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_dims
        Image dimensions. Inferred from inputs in None. (Default value = None)
    device
        device on which to create the array 'cuda:0', 'cuda:1', 'cpu' etc.
        Same as x if None. (Default value = None)

    Returns
    -------
    ret
        Optical flow from frame 1 to 2, projected to frame 1 epipolar line
        *[batch_shape,h,w,2]*

    """
    if batch_shape is None:
        batch_shape = flow.shape[:-3]

    if image_dims is None:
        image_dims = flow.shape[-3:-1]

    # shapes as list
    batch_shape = list(batch_shape)
    image_dims = list(image_dims)

    if uniform_pixel_coords is None:
        uniform_pixel_coords = ivy_svg.create_uniform_pixel_coords_image(
            image_dims, batch_shape, device
        )

    # BS x H x W x 3
    epipolar_lines = ivy_pg.transform(
        uniform_pixel_coords, fund_mat, batch_shape, image_dims
    )

    # BS x H x W x 2
    flow_pixels = uniform_pixel_coords[..., 0:2] + flow

    # BS x H x W x 1
    a = epipolar_lines[..., 0:1]
    b = epipolar_lines[..., 1:2]
    c = epipolar_lines[..., 2:3]

    x0 = flow_pixels[..., 0:1]
    y0 = flow_pixels[..., 1:2]

    bx0 = b * x0
    ay0 = a * y0
    a_sqrd = a**2
    b_sqrd = b**2
    ac = a * c
    bc = b * c
    a_sqrd_plus_b_sqrd = a_sqrd + b_sqrd

    x = (b * (bx0 - ay0) - ac) / (a_sqrd_plus_b_sqrd + MIN_DENOMINATOR)
    y = (a * (-bx0 + ay0) - bc) / (a_sqrd_plus_b_sqrd + MIN_DENOMINATOR)

    # BS x H x W x 2
    projected_pixels = ivy.concat((x, y), axis=-1)
    projected_flow = projected_pixels - uniform_pixel_coords[..., 0:2]
    return projected_flow


# noinspection PyUnresolvedReferences
def pixel_cost_volume(image1, image2, search_range, batch_shape=None):
    r"""Compute cost volume from image feature patch comparisons between first image
    :math:`\mathbf{X}_1\in\mathbb{R}^{h×w×d}` and second image :math:`\mathbf{
    X}_2\in\mathbb{R}^{h×w×d}`, as used in FlowNet paper.\n `[reference]
    <https://www.cv-foundation.org/openaccess/content_iccv_2015/papers
    /Dosovitskiy_FlowNet_Learning_Optical_ICCV_2015_paper.pdf>`_

    Parameters
    ----------
    image1
        Image 1 *[batch_shape,h,w,D]*
    image2
        Image 2 *[batch_shape,h,w,D]*
    search_range
        Search range for patch comparisons.
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)

    Returns
    -------
    ret
        Cost volume between the images *[batch_shape,h,w,(search_range*2+1)^2]*

    """
    if batch_shape is None:
        batch_shape = image1.shape[:-3]

    # shapes as list
    batch_shape = list(batch_shape)

    # shape info
    shape = image1.shape
    h = shape[-3]
    w = shape[-2]
    max_offset = search_range * 2 + 1

    # pad dims
    pad_dims = (
        [[0, 0]] * len(batch_shape) + [[search_range, search_range]] * 2 + [[0, 0]]
    )

    # BS x (H+2*SR) x (W+2*SR) x D
    padded_lvl = ivy.zero_pad(image2, pad_dims)

    # create list
    cost_vol = []

    # iterate through patches
    for y in range(0, max_offset):
        for x in range(0, max_offset):
            # BS x H x W x D
            tensor_slice = padded_lvl[..., y : y + h, x : x + w, :]

            # BS x H x W x 1
            cost = ivy.mean(image1 * tensor_slice, axis=-1, keepdims=True)

            # append to list
            cost_vol.append(cost)

    # BS x H x W x (max_offset^2)
    return ivy.concat(cost_vol, axis=-1)


# noinspection PyUnresolvedReferences
def velocity_from_flow_cam_coords_and_cam_mats(
    flow_t_to_tm1,
    cam_coords_t,
    cam_coords_tm1,
    cam_tm1_to_t_ext_mat,
    delta_t,
    uniform_pixel_coords=None,
    batch_shape=None,
    image_dims=None,
    device=None,
):
    """Compute relative cartesian velocity from optical flow, camera co-ordinates, and
    camera extrinsics.

    Parameters
    ----------
    flow_t_to_tm1
        Optical flow from frame t to t-1 *[batch_shape,h,w,2]*
    cam_coords_t
        Camera-centric homogeneous co-ordinates image in frame t *[batch_shape,h,w,4]*
    cam_coords_tm1
        Camera-centric homogeneous co-ordinates image in frame t-1 *[batch_shape,h,w,4]*
    cam_tm1_to_t_ext_mat
        Camera t-1 to camera t extrinsic projection matrix *[batch_shape,3,4]*
    delta_t
        Time difference between frame at timestep t-1 and t *[batch_shape,1]*
    uniform_pixel_coords
        Homogeneous uniform (integer) pixel co-ordinate images,
        inferred from image_dims if None *[batch_shape,h,w,3]* (Default value = None)
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_dims
        Image dimensions. Inferred from inputs in None. (Default value = None)
    device
        device on which to create the array 'cuda:0', 'cuda:1', 'cpu' etc.
        Same as x if None. (Default value = None)

    Returns
    -------
    ret
        Cartesian velocity measurements relative to the camera *[batch_shape,h,w,3]*

    """
    if batch_shape is None:
        batch_shape = flow_t_to_tm1.shape[:-3]

    if image_dims is None:
        image_dims = flow_t_to_tm1.shape[-3:-1]

    # shapes as list
    batch_shape = list(batch_shape)
    image_dims = list(image_dims)

    if device is None:
        device = ivy.dev(flow_t_to_tm1)

    if uniform_pixel_coords is None:
        uniform_pixel_coords = ivy_svg.create_uniform_pixel_coords_image(
            image_dims, batch_shape, device
        )

    # Interpolate cam coords from frame t-1

    # BS x H x W x 2
    warp = uniform_pixel_coords[..., 0:2] + flow_t_to_tm1

    # BS x H x W x 4
    cam_coords_tm1_interp = ivy_img.bilinear_resample(cam_coords_tm1, warp)

    # Project to frame t

    # BS x H x W x 4
    cam_coords_t_proj = ivy_tvg.cam_to_cam_coords(
        cam_coords_tm1_interp, cam_tm1_to_t_ext_mat, batch_shape, image_dims
    )

    # delta co-ordinates

    # BS x H x W x 3
    delta_cam_coords_t = (cam_coords_t - cam_coords_t_proj)[..., 0:3]

    # velocity

    # BS x H x W x 3
    vel = delta_cam_coords_t / ivy.reshape(delta_t, batch_shape + [1] * 3)

    # Validity mask

    # BS x H x W x 1
    validity_mask = (
        ivy.sum(
            ivy.astype(
                warp
                < ivy.array(
                    [image_dims[1], image_dims[0]], dtype="float32", device=device
                ),
                "int32",
            ),
            axis=-1,
            keepdims=True,
        )
        == 2
    )

    # pruned

    # BS x H x W x 3,    BS x H x W x 1
    return (
        ivy.where(validity_mask, vel, ivy.zeros_like(vel, device=device)),
        validity_mask,
    )


def project_cam_coords_with_object_transformations(
    cam_coords_1,
    id_image,
    obj_ids,
    obj_trans,
    cam_1_to_2_ext_mat,
    batch_shape=None,
    image_shape=None,
):
    """Compute velocity image from co-ordinate image, id image, and object
    transformations.

    Parameters
    ----------
    cam_coords_1
        Camera-centric homogeneous co-ordinates image in frame t
        *[batch_shape,image_shape,4]*
    id_image
        Image containing per-pixel object ids *[batch_shape,h,w,1]*
    obj_ids
        Object ids *[batch_shape,num_obj,1]*
    obj_trans
        Object transformations for this frame over time *[batch_shape,num_obj,3,4]*
    cam_1_to_2_ext_mat
        Camera 1 to camera 2 extrinsic projection matrix *[batch_shape,3,4]*
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_shape
        Image dimensions. Inferred from inputs in None. (Default value = None)

    Returns
    -------
    ret
        Relative velocity image *[batch_shape,image_shape,3]*

    """
    if batch_shape is None:
        batch_shape = cam_1_to_2_ext_mat.shape[:-2]
    num_batch_dims = len(batch_shape)

    if image_shape is None:
        image_shape = cam_coords_1.shape[num_batch_dims:-1]
    num_image_dims = len(image_shape)

    # shapes as list
    batch_shape = list(batch_shape)
    image_shape = list(image_shape)
    num_batch_dims = len(batch_shape)

    # Transform the co-ordinate image by each transformation

    # BS x (num_obj x 3) x 4
    obj_trans = ivy.reshape(obj_trans, batch_shape + [-1, 4])

    # BS x 4 x IS
    cam_coords_1_ = ivy.permute_dims(
        cam_coords_1,
        axes=list(range(num_batch_dims))
        + [
            i + num_batch_dims for i in ([num_image_dims] + list(range(num_image_dims)))
        ],
    )

    # BS x 4 x (IS)
    cam_coords_1_ = ivy.reshape(cam_coords_1_, batch_shape + [4, -1])

    # BS x (num_obj x 3) x (IS)
    cam_coords_2_all_obj_trans = ivy.matmul(obj_trans, cam_coords_1_)

    # BS x (IS) x (num_obj x 3)
    cam_coords_2_all_obj_trans = ivy.swapaxes(cam_coords_2_all_obj_trans, -1, -2)

    # BS x IS x num_obj x 3
    cam_coords_2_all_obj_trans = ivy.reshape(
        cam_coords_2_all_obj_trans, batch_shape + image_shape + [-1, 3]
    )

    # Multiplier

    # BS x [1]*NID x num_obj
    obj_ids = ivy.reshape(obj_ids, batch_shape + [1] * num_image_dims + [-1])

    # BS x IS x num_obj x 1
    multiplier = ivy.astype(ivy.expand_dims(obj_ids == id_image, axis=-1), "float32")

    # compute validity mask, for pixels which are on moving objects

    # BS x IS x 1
    motion_mask = ivy.sum(multiplier, axis=-2) > 0

    # make invalid transformations equal to zero

    # BS x IS x num_obj x 3
    cam_coords_2_all_obj_trans_w_zeros = cam_coords_2_all_obj_trans * multiplier

    # reduce to get only valid transformations

    # BS x IS x 3
    cam_coords_2_all_obj_trans = ivy.sum(cam_coords_2_all_obj_trans_w_zeros, axis=-2)

    # find cam coords to for zero motion pixels

    # BS x IS x 3
    cam_coords_2_wo_motion = ivy_tvg.cam_to_cam_coords(
        cam_coords_1, cam_1_to_2_ext_mat, batch_shape, image_shape
    )

    # BS x IS x 4
    cam_coords_2_all_trans_homo = ivy_mech.make_coordinates_homogeneous(
        cam_coords_2_all_obj_trans, batch_shape + image_shape
    )
    cam_coords_2 = ivy.where(
        motion_mask, cam_coords_2_all_trans_homo, cam_coords_2_wo_motion
    )

    # return

    # BS x IS x 3,    BS x IS x 1
    return cam_coords_2, motion_mask


def velocity_from_cam_coords_id_image_and_object_trans(
    cam_coords_t,
    id_image,
    obj_ids,
    obj_trans,
    delta_t,
    batch_shape=None,
    image_shape=None,
    device=None,
):
    """Compute velocity image from co-ordinate image, id image, and object
    transformations.

    Parameters
    ----------
    cam_coords_t
        Camera-centric homogeneous co-ordinates image in frame t
        *[batch_shape,image_shape,4]*
    id_image
        Image containing per-pixel object ids *[batch_shape,image_shape,1]*
    obj_ids
        Object ids *[batch_shape,num_obj,1]*
    obj_trans
        Object transformations for this frame over time *[batch_shape,num_obj,3,4]*
    delta_t
        Time difference between frame at timestep t-1 and t *[batch_shape,1]*
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_shape
        Image dimensions. Inferred from inputs in None. (Default value = None)
    device
        device on which to create the array 'cuda:0', 'cuda:1', 'cpu' etc.
        Same as x if None. (Default value = None)

    Returns
    -------
    ret
        Relative velocity image *[batch_shape,image_shape,3]*

    """
    if batch_shape is None:
        batch_shape = delta_t.shape[:-1]
    num_batch_dims = len(batch_shape)

    if image_shape is None:
        image_shape = cam_coords_t.shape[num_batch_dims:-1]

    if device is None:
        device = ivy.dev(cam_coords_t)

    # shapes as list
    batch_shape = list(batch_shape)
    image_shape = list(image_shape)

    # get co-ordinate re-projections

    # BS x IS x 4
    (
        cam_coords_t_all_trans,
        motion_mask,
    ) = project_cam_coords_with_object_transformations(
        cam_coords_t,
        id_image,
        obj_ids,
        obj_trans,
        ivy.eye(4, batch_shape=batch_shape)[..., 0:3, :],
        batch_shape,
        image_shape,
    )

    # BS x IS x 4
    cam_coords_t_all_trans = ivy.where(
        motion_mask,
        cam_coords_t_all_trans,
        ivy.zeros_like(cam_coords_t_all_trans, device=device),
    )

    # compute velocities

    # BS x IS x 3
    vel = (cam_coords_t[..., 0:3] - cam_coords_t_all_trans[..., 0:3]) / delta_t

    # prune velocities

    # BS x IS x 3
    return ivy.where(motion_mask, vel, ivy.zeros_like(vel, device=device))


def flow_from_cam_coords_id_image_and_object_trans(
    cam_coords_f1,
    id_image,
    obj_ids,
    obj_trans,
    calib_mat,
    cam_1_to_2_ext_mat,
    batch_shape=None,
    image_shape=None,
):
    """Compute optical flow from co-ordinate image, id image, and object
    transformations.

    Parameters
    ----------
    cam_coords_f1
        Camera-centric homogeneous co-ordinates image in frame t
        *[batch_shape,image_shape,4]*
    id_image
        Image containing per-pixel object ids *[batch_shape,image_shape,1]*
    obj_ids
        Object ids *[batch_shape,num_obj,1]*
    obj_trans
        Object transformations for this frame over time *[batch_shape,num_obj,3,4]*
    calib_mat
        Calibration matrix *[batch_shape,3,3]*
    cam_1_to_2_ext_mat
        Camera 1 to camera 2 extrinsic projection matrix *[batch_shape,3,4]*
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_shape
        Image dimensions. Inferred from inputs in None. (Default value = None)

    Returns
    -------
    ret
        Relative velocity image *[batch_shape,image_shape,3]*

    """
    if batch_shape is None:
        batch_shape = calib_mat.shape[:-2]
    num_batch_dims = len(batch_shape)

    if image_shape is None:
        image_shape = cam_coords_f1.shape[num_batch_dims:-1]

    # shapes as list
    batch_shape = list(batch_shape)
    image_shape = list(image_shape)

    # get co-ordinate re-projections

    # BS x IS x 3
    cam_coords_trans_f2, _ = project_cam_coords_with_object_transformations(
        cam_coords_f1,
        id_image,
        obj_ids,
        obj_trans,
        cam_1_to_2_ext_mat,
        batch_shape,
        image_shape,
    )

    # co-ordinates to pixel co-ordinates

    # BS x IS x 3
    pixel_coords_f1 = ivy_svg.cam_to_ds_pixel_coords(
        cam_coords_f1, calib_mat, batch_shape, image_shape
    )
    pixel_coords_trans_f2 = ivy_svg.cam_to_ds_pixel_coords(
        cam_coords_trans_f2, calib_mat, batch_shape, image_shape
    )

    # unscaled pixel coords
    unscaled_pixel_coords_f1 = pixel_coords_f1[..., 0:2] / (
        pixel_coords_f1[..., -1:] + MIN_DENOMINATOR
    )
    unscaled_pixel_coords_f2 = pixel_coords_trans_f2[..., 0:2] / (
        pixel_coords_trans_f2[..., -1:] + MIN_DENOMINATOR
    )

    # optical flow

    # BS x IS x 2
    return unscaled_pixel_coords_f2 - unscaled_pixel_coords_f1
